"""
The result of measurement validation.
"""

from __future__ import annotations

import abc
from copy import deepcopy
from typing import List, Optional

from mlte.measurement.evaluation.evalution_result import EvaluationResult


def _has_callable(type, name) -> bool:
    """Determine if `type` has a callable attribute with the given name."""
    return hasattr(type, name) and callable(getattr(type, name))


# -----------------------------------------------------------------------------
# Measurement Binding
# -----------------------------------------------------------------------------


class Binding(metaclass=abc.ABCMeta):
    """The base class for measurement bindings."""

    @classmethod
    def __subclasshook__(cls, subclass):
        """Define the interface for all concrete ValidationResult."""
        return all(_has_callable(subclass, method) for method in ["__bool__"])

    def __init__(self):
        """
        Initialize a Binding instance.
        """
        self.measurement_name = ""
        """The name of the measurement that is bound."""
        self.property_names: List[str] = []
        """The names of the properties to which measurement is bound."""

        raise NotImplementedError("Cannot instantiate abstract Binding")

    def __bool__(self) -> bool:
        """Implicit boolean conversion."""
        raise NotImplementedError(
            "Boolean conversion of abstract Binding is ambiguous."
        )

    def __eq__(self, other: object) -> bool:
        """Equality comparison."""
        if not isinstance(other, Binding):
            raise NotImplementedError(
                f"Equality on Binding and {type(other).__name__}"
            )
        return self.measurement_name == other.measurement_name and set(
            self.property_names
        ) == set(other.property_names)

    def __neq__(self, other: object) -> bool:
        """Inequality comparison."""
        return not (self == other)


class Bound(Binding):
    """A Bound instance represents a binding from measurement to property."""

    def __init__(self, measurement, *properties: str):
        """
        Initialize a Bound instance.

        :param measurement: The Measurement to which binding is attached
        :type measurement: Measurement
        :param properties: The names of the properties
        to which the measurement is bound
        :type properties: str
        """
        self.measurement_name = measurement.name
        """The name of the measurement that is bound."""
        self.property_names: List[str] = [name for name in properties]
        """The names of the properties to which measurement is bound."""

    def __bool__(str) -> bool:
        """Implicit boolean conversion."""
        return True

    def __str__(self) -> str:
        """Return a string representation of Bound."""
        property_names = ",".join(name for name in self.property_names)
        return f"Bound: {self.measurement_name} -> {property_names}"


class Unbound(Binding):
    """An Unbound instand represents an unbound measurement."""

    def __init__(self):
        """
        Initialize a Unbound instance.
        """
        self.measurement_name = ""
        """The name of the measurement that is bound."""
        self.property_names: List[str] = []
        """The names of the properties to which measurement is bound."""

    def __bool__(self) -> bool:
        """Implicit boolean conversion."""
        return False

    def __str__(self) -> str:
        """Return a string representation of Unbound."""
        return "Unbound"


# -----------------------------------------------------------------------------
# Validation Results
# -----------------------------------------------------------------------------


class ValidationResult(metaclass=abc.ABCMeta):
    """The base class for measurement validation results."""

    @classmethod
    def __subclasshook__(cls, subclass):
        """Define the interface for all concrete ValidationResult."""
        return all(
            _has_callable(subclass, method)
            for method in ["__bool__", "__str__"]
        )

    def __init__(self):
        """
        Initialize a ValidationResult instance.
        """

        self.validator_name = ""
        """The name of the validator that produced the result."""

        self.data: Optional[EvaluationResult] = None
        """
        The EvaluationResult on which the Validator
        that produced this ValidationResult was invoked.
        """

        self.binding: Binding = Unbound()
        """The Binding for the ValidationResult."""

        self.message = ""
        """The message indicating the reason for status."""

    def _from_validator(self, validator) -> ValidationResult:
        """
        Set the `validator_name` field of the ValidationResult
        to indicate the validator that produced the result.

        :param validator: The Validator instance that produced the result
        :type validator: Validator

        :return: The ValidationResult instance (`self`)
        :rtype: ValidationResult
        """
        self.validator_name = validator.name
        return self

    def _from_data(self, data: EvaluationResult) -> ValidationResult:
        """
        Set the `data` field of the ValidationResult
        to indicate the EvaluationResult from which
        it was generated.

        :param data: The EvaluationResult on which the
        Validator that produced this instance was invoked
        :type data: EvaluationResult

        :return: The ValidationResult instance (`self`)
        :rtype: ValidationResult
        """
        # TODO(Kyle): This is probably not necessary,
        # and is certainly overkill for the current
        # measurements that we have implemented. Revisit.
        self.data = deepcopy(data)
        return self

    def _with_binding(self, binding: Binding) -> ValidationResult:
        """
        Set the `binding` field of the ValidationResult.

        :param binding: The Binding instance
        :type binding: Binding

        :return: The ValidationResult instance (`self`)
        :rtype: ValidationResult
        """
        self.binding = binding
        return self

    def _is_bound(self) -> bool:
        """
        Determine if the ValidationResult is bound to a property.

        :return: `True` if the result is bound, `False` otherwise
        :rtype: bool
        """
        return bool(self.binding)

    def __eq__(self, other: object) -> bool:
        """Equality comparison."""
        if not isinstance(other, ValidationResult):
            raise NotImplementedError(
                f"Equality on ValidationResult and {type(other).__name__}"
            )
        return (
            self.validator_name == other.validator_name
            and self.binding == other.binding
        )

    def __neq__(self, other: object) -> bool:
        """Inequality comparison."""
        return not (self == other)


class Success(ValidationResult):
    """Indicates successful measurement validation."""

    def __init__(self, message: str = ""):
        """
        Initialize a Success validation result instance.

        :param validator_name: The name of the validator
        :type validator_name: str
        :param message: Optional message
        :type message: str
        """
        super().__init__()

        self.message = message
        """The message indicating the reason for success."""

    def __bool__(self) -> bool:
        """Implicit boolean conversion."""
        return True

    def __str__(self) -> str:
        """String representation."""
        return "Success"


class Failure(ValidationResult):
    """Indicates failed measurement validation."""

    def __init__(self, message: str = ""):
        """
        Initialize a Failure validation result instance.

        :param validator_name: The name of the validator
        :type validator_name: str
        :param message: Optional message
        :type message: str
        """
        super().__init__()

        self.message = message
        """The message indicating the reason for failure."""

    def __bool__(self) -> bool:
        """Implicit boolean conversion."""
        return False

    def __str__(self) -> str:
        """String representation."""
        return "Failure"


class Ignore(ValidationResult):
    """Indicates ignored measurement validation."""

    def __init__(self, message: str):
        """
        Initialize an Ignore validatation result instance.

        :param message: Message indicating the reason validation is ignored
        :type message: str
        """
        super().__init__()

        self.message = message
        """The message indicating the reason validation is ignored."""

    def __bool__(self) -> bool:
        """Implicit boolean conversion."""
        raise RuntimeError("Boolean conversion for Ignore() is ambiguous.")

    def __str__(self) -> str:
        """String representation."""
        return "Ignore"



"""
The interface for measurement validation.
"""

import typing
from typing import Callable
from ..evaluation import EvaluationResult
from ..validation import ValidationResult


class Validator:
    """
    The Validator class defines the interface for measurement validators.
    """

    @typing.no_type_check
    def __init__(
        self,
        name: str,
        callback: Callable[[EvaluationResult], ValidationResult],
    ):
        """
        Initialize a Validator instance.

        :param name: The validator identifier
        :type name: str
        :param callback: The callable that implements validation
        :type callback: Callable[[EvaluationResult], ValidationResult]
        """
        self.name: str = name
        """The human-readable identifier for the Validator."""

        self.callback: Callable[[EvaluationResult], ValidationResult] = callback
        """The callback that implements validation."""

    def __call__(self, result: EvaluationResult) -> ValidationResult:
        """
        Invoke the validation callback

        :param result: The result of measurement evaluation
        :type result: EvaluationResult

        :return: The result of measurement validation
        :rtype: ValidationResult
        """
        return self.callback(result)



from .report import (
    Dataset,
    User,
    UseCase,
    Limitation,
    Metadata,
    ModelDetails,
    ModelSpecification,
    Considerations,
    Report,
)

from .render import render

__all__ = [
    "Dataset",
    "User",
    "UseCase",
    "Limitation",
    "Metadata",
    "ModelDetails",
    "ModelSpecification",
    "Considerations",
    "Report",
    "render",
]



"""
The superclass for all model properties.
"""

from __future__ import annotations

import abc
import importlib
from typing import Dict, Type

# The names of the properties submodules
SUBMODULES = ["costs"]


def _has_callable(type, name) -> bool:
    """Determine if `type` has a callable attribute with the given name."""
    return hasattr(type, name) and callable(getattr(type, name))


class Property(metaclass=abc.ABCMeta):
    """The Property type represents an abstract model property."""

    @classmethod
    def __subclasshook__(cls, subclass):
        """Define the interface for all concrete properties."""
        return all(
            _has_callable(subclass, method)
            for method in ["__init__", "__repr__"]
        )

    def __init__(self, name: str, description: str):
        """
        Initialize a Property instance.

        :param name: The name of the property
        :type name: str
        :param description: The description of the property
        :type description: str
        """
        self.name: str = name
        """The name of the property."""
        self.description = description
        """The description of the property."""

    def _to_document(self) -> Dict[str, str]:
        """
        Save a Property instance to a document.

        :return: The document
        :rtype: Dict[str, str]
        """
        return {
            "name": self.name,
            "repr": repr(self),
            "description": self.description,
        }

    @staticmethod
    def _from_document(document: Dict[str, str]) -> Property:
        """
        Load a Property instance from a document.

        :param document: The document for the saved propery
        :type document: Dict[str, str]

        :return: The loaded property
        :rtype: Property
        """
        return _load_from_document(document)


def _get_class_name(property_repr: str) -> str:
    """
    Extract the class name from the property representation.

    :param property_repr: The representation
    :type property_repr: str

    :return: The class name
    :rtype: str
    """
    return property_repr[: property_repr.index("(")]


def _load_from_document(document: Dict[str, str]) -> Property:
    """
    Load a Property instance from its identifier.

    :param name: The name of the property
    :type name: str

    :return: The loaded property
    :rtype: Property
    """
    if "name" not in document or "repr" not in document:
        raise RuntimeError("Saved property is malformed.")
    property_repr = document["repr"]

    # Extract the classname from the call
    classname = _get_class_name(property_repr)

    # Load the class type from the module
    for submodule in SUBMODULES:
        module = importlib.import_module(
            f".{submodule}", package="mlte.property"
        )
        try:
            class_: Type[Property] = getattr(module, classname)
        except AttributeError:
            continue

        # Instantiate the property
        return class_()  # type: ignore

    raise RuntimeError(f"Property {document['name']} not found")



"""
Utilities for execution of measurements.
"""

import concurrent.futures
from typing import Callable, Any, List


def concurrently(*callables: Callable[[], Any]) -> List[Any]:
    """
    Run an arbitrary number of functions concurrently.

    :param callables: The callables to execute
    :type callables: Callable[[], Any]

    :return: The results of all callables, in
    the order in which they were passed to call
    :rtype: List[Any]
    """
    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = [executor.submit(callable) for callable in callables]
        return [future.result() for future in futures]



"""
Storage capacity measurement for locally-stored objects.
"""

import os
from typing import Dict, Any

from ..measurement import Measurement
from ..evaluation import Integer
from ..validation import Validator, Success, Failure


class LocalObjectSize(Measurement):
    """Measure the size of a locally-stored object."""

    def __init__(self):
        """Initialize a new LocalObjectSize measurement."""
        super().__init__("LocalObjectSize")

    def __call__(self, path: str) -> Dict[str, Any]:
        """
        Compute the size of the object at `path`.

        :param path: The path to the object
        :type path: str

        :return: The size of the object, in bytes
        :rtype: Dict
        """
        if not os.path.isfile(path) and not os.path.isdir(path):
            raise RuntimeError(f"Invalid path: {path}")

        # If the object is just a file, return it immediately
        if os.path.isfile(path):
            return {"total_size": os.path.getsize(path)}

        # Otherwise, the object must be directory
        assert os.path.isdir(path), "Broken invariant."

        total_size = 0
        for dirpath, _, filenames in os.walk(path):
            for name in filenames:
                path = os.path.join(dirpath, name)
                if not os.path.islink(path):
                    total_size += os.path.getsize(path)

        return {"total_size": total_size}

    def semantics(self, data: Dict[str, Any]) -> Integer:
        """Provide semantics for measurement output."""
        assert "total_size" in data, "Broken invariant."
        return Integer(self, data["total_size"])

    def with_validator_size_not_greater_than(
        self, threshold: int
    ) -> Measurement:
        """
        Add a validator for the size of the object.

        :param threshold: The threshold for object size, in bytes
        :type threshold: int

        :return: The measurement instance (`self`)
        :rtype: Measurement
        """
        return self.with_validator(
            Validator(
                "MaximumSize",
                lambda size: Success(
                    f"Object size {size} below threshold {threshold}"
                )
                if size.value <= threshold
                else Failure(
                    f"Object size {size} exceeds threshold {threshold}"
                ),
            )
        )



from .local_process_memory_consumption import (
    LocalProcessMemoryConsumption,
    MemoryStatistics,
)

# TODO(Kyle): Find a more elegant way to do this
__all__ = ["LocalProcessMemoryConsumption", "MemoryStatistics"]



"""
An EvaluationResult instance for a scalar, real value.
"""

from .evalution_result import EvaluationResult


class Real(EvaluationResult):
    """
    Real implements the EvaluationResult
    interface for a single real value.
    """

    def __init__(self, measurement, value: float):
        """
        Initialize a Real instance.

        :param measurement: The generating measurement
        :type measurement: Measurement
        :param value: The real value
        :type value: float
        """
        assert isinstance(value, float), "Argument must be `float`."

        super().__init__(measurement)

        self.value = value
        """The wrapped real value."""

    def __str__(self) -> str:
        """Return a string representation of the Real."""
        return f"{self.value}"



"""
A collection of properties and their measurements.
"""

from __future__ import annotations

import os
import time
import json
from itertools import groupby, combinations
from typing import List, Dict, Iterable, Any

from ..property import Property
from ..measurement.validation import ValidationResult
from .._private.schema import SUITE_LATEST_SCHEMA_VERSION


def _unique(collection: List[str]) -> bool:
    """
    Determine if all elements of a collection are unique.

    :param collection: The collection
    :type collection: Iterable

    :return: `True` if all elements are unique, `False` otherwise
    :rtype: bool
    """
    return len(set(collection)) == len(collection)


def _all_equal(iterable: Iterable[Any]) -> bool:
    """
    Determine if all elements of an iterable are equivalent.

    :param iterable: The iterable
    :type iterable: Iterable[Any]

    :return: `True` if all elements are equal, `False` otherwise
    :rtype: bool
    """
    g = groupby(iterable)
    return next(g, True) and not next(g, False)  # type: ignore


class SuiteReport:
    """SuiteReport represents the result of collecting a Suite."""

    def __init__(self, document: Dict[str, Any]):
        """
        Initialize a SuiteReport instance.

        :param document: The data produced by the Suite
        :type document: Dict[str, Any]]
        """
        self.document = document
        """The document produced by the Suite."""


class Suite:
    """
    The Suite class integrates properties, measurements,
    and the results of measurement evaluation and validation.
    """

    def __init__(self, name: str, *properties: Property):
        """
        Initialize a Suite instance.

        :param name: The identifier for the suite
        :type name: str
        :param properties: The collection of properties that compose the suite
        :type properties: Property
        """
        # TODO(Kyle): What additional metadata should
        # we store at the level of a Suite?

        if not isinstance(name, str):
            raise RuntimeError(f"Invalid name for Suite: {name}")

        self.name = name
        """The human-readable identifier for the Suite."""

        self.properties = [p for p in properties]
        """The collection of properties that compose the Suite."""

        if not _unique([p.name for p in self.properties]):
            raise RuntimeError("All properties in Suite must be unique.")

    # -------------------------------------------------------------------------
    # Property Manipulation
    # -------------------------------------------------------------------------

    def add_property(self, property: Property):
        """
        Add a property to the suite.

        :param property: The property to add
        :type property: Property
        """
        if property.name in (p.name for p in self.properties):
            raise RuntimeError("Properties in Suite must be unique.")
        self.properties.append(property)

    def has_property(self, name: str) -> bool:
        """
        Determine if the suite contains a particular property.

        :param name: The name of the property
        :type name: str

        :return: `True` if the suite has the property, `False` otherwise
        :rtype: bool
        """
        return any(property.name == name for property in self.properties)

    def get_property(self, name: str) -> Property:
        """
        Get the property with the given identifier, if present.

        :param name: The name of the property
        :type name: str

        :return: The property identified by `name`
        :rtype: Property
        """
        if not self.has_property(name):
            raise RuntimeError(f"Property {name} not found")

        for property in self.properties:
            if property.name == name:
                return property

        raise RuntimeError("Unreachable")

    # -------------------------------------------------------------------------
    # Save / Load
    # -------------------------------------------------------------------------

    def save(self, path: str):
        """
        Save the Suite to `path`.

        :param path: The path to which the Suite is saved
        :type path: str
        """
        # TODO(Kyle): Implement this.
        if path.startswith("s3://"):
            raise NotImplementedError("Save as S3 object not implemented.")

        document = {
            "name": self.name,
            "properties": [
                property._to_document() for property in self.properties
            ],
        }
        with open(path, "w") as f:
            json.dump(document, f)

    @staticmethod
    def from_file(path: str) -> Suite:
        """
        Load a Suite instance from file.

        :param path: The path to the saved Suite
        :type path: str

        :return: The loaded Suite
        :rtype: Suite
        """
        if not os.path.exists(path):
            raise RuntimeError(f"Suite does not exist at path {path}")
        if not os.path.isfile(path):
            raise RuntimeError(f"Suite at path {path} is not a file")

        with open(path, "r") as f:
            document = json.load(f)

        if "name" not in document:
            raise RuntimeError(f"Suite at path {path} missing 'name'")
        if "properties" not in document:
            raise RuntimeError(f"Suite at path {path} missing 'properties'")
        if not isinstance(document["properties"], list):
            raise RuntimeError(f"Suite at path {path} is corrupt")

        suite = Suite(document["name"])
        for pdoc in document["properties"]:
            suite.add_property(Property._from_document(pdoc))

        return suite

    # -------------------------------------------------------------------------
    # Report Generation
    # -------------------------------------------------------------------------

    def collect(
        self, *results: ValidationResult, strict: bool = True
    ) -> SuiteReport:
        """
        Collect validation results from all measurements,
        and generate a suite report from these measurements.

        If the `strict` flag is set to `True`, then all properties
        declared in the suite must have at least one (1) measurement
        bound to them in order to proceed with report generation.
        Suite collection is set to `strict` mode by default.

        :param results: Validation results
        :type results: ValidationResult
        :param strict: Flag indicating strict measurement requirements
        :type strict: bool

        :return: The suite report
        :rtype: SuiteReport
        """
        self._validate_uniqueness(*results)
        self._validate_bindings(*results)
        if strict:
            self._validate_property_coverage(*results)

        properties = [
            self._collect_property(property, *results)
            for property in self.properties
        ]
        document: Dict[str, Any] = {
            "schema_version": SUITE_LATEST_SCHEMA_VERSION,
            "name": self.name,
            "timestamp": f"{int(time.time())}",
            "properties": properties,
        }
        return SuiteReport(document)

    def _validate_uniqueness(self, *results: ValidationResult):
        """
        Validate the uniqueness of validation results.

        :param results: The validation results
        :type results: ValidationResult

        :raises RuntimeError: If any two results are equivalent
        """
        for pair in combinations(results, 2):
            if pair[0] == pair[1]:
                raise RuntimeError("All validation results must be unique")

    def _validate_bindings(self, *results: ValidationResult):
        """
        Validate bindings.

        :param results: Validation results
        :type results: ValidationResult

        :raises RuntimeError: If any validation result is not bound
        :raises RuntimeError: If any validation result is bound
        to a property that does not exist in the suite
        """
        if not all(r._is_bound() for r in results):
            raise RuntimeError("All ValidationResult must be bound.")
        _ = all(self._validate_bindings_for_result(r) for r in results)

    def _validate_bindings_for_result(self, result: ValidationResult):
        """
        Validate the bindings for an individual ValidationResult.

        :param result: The validation result
        :type result: ValidationResult

        :raises RuntimeError: If any validation result is bound
        to a property that does not exist in the suite
        """
        property_names = [p.name for p in self.properties]
        if not all(
            name in property_names for name in result.binding.property_names
        ):
            raise RuntimeError(
                (
                    f"Result from validator {result.validator_name} "
                    "bound to nonexistent property."
                )
            )

    def _validate_property_coverage(self, *results: ValidationResult):
        """
        Ensure that every property in the suite has at
        least one associated measurement in `results`.

        :param results: The validation results
        :type results: ValidationResult

        :raises RuntimeError: If any property is uncovered
        """
        for property in self.properties:
            if not any(
                property.name in r.binding.property_names for r in results
            ):
                raise RuntimeError(
                    f"Property {property.name} has no bound measurements"
                )

    def _collect_property(
        self, property: Property, *results: ValidationResult
    ) -> Dict[str, Any]:
        """
        Collect the results for an individual property
        from result set into a property-level document.

        :param property: The property of interest
        :type property: Property
        :param results: The result collection
        :type results: ValidationResult

        :return: The property-level document
        :rtype: Dict[str, Any]
        """
        # Filter results relevant to property
        results_for_property = [
            r for r in results if property.name in r.binding.property_names
        ]
        measurements = []
        for _, group in groupby(
            results_for_property, key=lambda vr: vr.binding.measurement_name
        ):
            measurements.append(
                self._collect_measurement(*(vr for vr in group))
            )

        document = {
            "name": property.name,
            "description": property.description,
            "measurements": measurements,
        }
        return document

    def _collect_measurement(
        self, *results: ValidationResult
    ) -> Dict[str, Any]:
        """
        Collect results into a measurement-level document.

        :param result: The validation results for the measurement
        :type results: ValidationResult

        :return: The measurement-level document
        :rtype: Dict[str, Any]
        """
        assert len(results) > 0, "Broken invariant."
        assert _all_equal(
            result.binding.measurement_name for result in results
        ), "Broken invariant."
        measurement_name = results[0].binding.measurement_name
        document = {
            "name": measurement_name,
            "validators": [
                {
                    "name": vr.validator_name,
                    "result": f"{vr}",
                    "message": vr.message,
                }
                for vr in results
            ],
        }
        return document



"""
TrainingComputeCost property definition.
"""

from ..property import Property
from ..._private.text import cleantext


class TrainingComputeCost(Property):
    """
    The TrainingComputeCost property reflects the costs
    of model training associated with compute resources.
    """

    def __init__(self):
        """
        Initialize a TrainingComputeCost instance.
        """
        super().__init__(
            "TrainingComputeCost",
            cleantext(
                """
                The TrainingComputeCost property assesses the
                computational requirements of model training.
                This might be measured in terms of CPU utilization
                for training processes that run locally, or the cost
                of compute resources required for training processes
                that run on on-demand cloud infrastructure.
                """
            ),
        )

    def __repr__(self) -> str:
        """Return the representation needed to reconstruct the object."""
        return f"{self.name}()"



"""
Memory consumption measurement for local training processes.
"""

import time
import subprocess
from typing import Dict, Any

from ..measurement import Measurement
from ..evaluation import EvaluationResult
from ..validation import Validator, Success, Failure
from ..._private.platform import is_windows


class MemoryStatistics(EvaluationResult):
    """
    The MemoryStatistics class encapsulates data
    and functionality for tracking and updating memory
    consumption statistics for a running process.
    """

    def __init__(self, measurement: Measurement, avg: int, min: int, max: int):
        """
        Initialize a MemoryStatistics instance.

        :param measurement: The generating measurement
        :type measurement: Measurement
        :param avg: The average memory consumption
        :type avg: int
        :param min: The minimum memory consumption
        :type avg: int
        :param max: The maximum memory consumption
        :type max: int
        """
        super().__init__(measurement)

        self.avg = avg
        """The average memory consumption (KB)."""

        self.min = min
        """The minimum memory consumption (KB)."""

        self.max = max
        """The maximum memory consumption (KB)."""

    def __str__(self) -> str:
        """Return a string representation of MemoryStatistics."""
        s = ""
        s += f"Average: {self.avg}\n"
        s += f"Minimum: {self.min}\n"
        s += f"Maximum: {self.max}"
        return s


def _get_memory_usage(pid: int) -> int:
    """
    Get the current memory usage for the process with `pid`.

    :param pid: The identifier of the process
    :type pid: int

    :return: The current memory usage in KB
    :rtype: int
    """
    # sudo pmap 917 | tail -n 1 | awk '/[0-9]K/{print $2}'
    try:
        with subprocess.Popen(
            ["pmap", f"{pid}"],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        ) as pmap, subprocess.Popen(
            ["tail", "-n", "1"],
            stdin=pmap.stdout,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        ) as tail:
            used = subprocess.check_output(
                ["awk", "/[0-9]K/{print $2}"],
                stdin=tail.stdout,
                stderr=subprocess.DEVNULL,
            )
        return int(used.decode("utf-8").strip()[:-1])
    except ValueError:
        return 0


class LocalProcessMemoryConsumption(Measurement):
    """Measure memory consumption for a local training process."""

    def __init__(self):
        """Initialize a LocalProcessMemoryConsumption instance."""
        super().__init__("LocalProcessMemoryConsumption")
        if is_windows():
            raise RuntimeError(
                f"Measurement {self.name} is not supported on Windows."
            )

    def __call__(self, pid: int, poll_interval: int = 1) -> Dict[str, Any]:
        """
        Monitor memory consumption of process at `pid` until exit.

        :param pid: The process identifier
        :type pid: int
        :param poll_interval: The poll interval, in seconds
        :type poll_interval: int

        :return The collection of memory usage statistics
        :rtype: Dict
        """
        stats = []
        while True:
            kb = _get_memory_usage(pid)
            if kb == 0:
                break
            stats.append(kb)
            time.sleep(poll_interval)

        return {
            "avg_consumption": int(sum(stats) / len(stats)),
            "min_consumption": min(stats),
            "max_consumption": max(stats),
        }

    def semantics(self, data: Dict[str, Any]) -> MemoryStatistics:
        """
        Provide semantics for measurement output.

        :param data: Measurement output data
        :type data: Dict

        :return: Memory consumption statistics
        :rtype: MemoryStatistics
        """
        assert "avg_consumption" in data, "Broken invariant."
        assert "min_consumption" in data, "Broken invariant."
        assert "max_consumption" in data, "Broken invariant."
        return MemoryStatistics(
            self,
            avg=data["avg_consumption"],
            min=data["min_consumption"],
            max=data["max_consumption"],
        )

    def with_validator_max_consumption_not_greater_than(
        self, threshold: int
    ) -> Measurement:
        """
        Add a validator for maximum memory consumption.

        :param threshold: The threshold value for maximum consumption
        :type threshold: int

        :return: The measurement instance (`self`)
        :rtype: Measurement
        """
        return self.with_validator(
            Validator(
                "MaximumConsumption",
                lambda stats: Success(
                    f"Maximum consumption {stats.max} "
                    f"below threshold {threshold}"
                )
                if stats.max <= threshold
                else Failure(
                    (
                        f"Maximum consumption {stats.max} "
                        f"exceeds threshold {threshold}"
                    )
                ),
            )
        )

    def with_validator_avg_consumption_not_greater_than(
        self, threshold: float
    ) -> Measurement:
        """
        Add a validator for average memory consumption.

        :param threshold: The threshold value for average consumption
        :type threshold: int

        :return: The measurement instance (`self`)
        :rtype: Measurement
        """
        return self.with_validator(
            Validator(
                "AverageConsumption",
                lambda stats: Success(
                    f"Average consumption {stats.avg} "
                    f"below threshold {threshold}"
                )
                if stats.avg <= threshold
                else Failure(
                    (
                        f"Average consumption {stats.avg} "
                        f"exceeds threshold {threshold}"
                    )
                ),
            )
        )



"""
Utilities for report rendering.
"""

import os
import tempfile
import webbrowser
from typing import Union, Optional

from .report import Report
from .html import _connected


def render(target: Union[Report, str]):
    """
    Render a Report as a web page.

    If possible, the rendered report is automatically
    opened in an available window in the default browser.

    The argument `target` may be specified as either the
    Report instance itself, or the HTML content generated
    via a call to Report.to_html(). Rendering reports from
    the local filesystem is currently unsupported.

    :param target: The report content to render
    :type target: Union[Report, str]
    """
    # TODO(Kyle): Validate HTML input.

    if not _connected():
        raise RuntimeError(
            "An internet connection is required to render a report."
        )

    html: Optional[str] = (
        target.to_html() if isinstance(target, Report) else target
    )
    assert html is not None, "Broken invariant."

    with tempfile.TemporaryDirectory() as dirname:
        path = os.path.join(dirname, "index.html")
        with open(path, "w") as f:
            f.write(html)
        webbrowser.open(f"file://{os.path.realpath(path)}")



"""
TaskEfficacy property definition.
"""

from ..property import Property
from ..._private.text import cleantext


class TaskEfficacy(Property):
    """
    The TaskEfficacy property assesses the ability of
    a model to perform the task for which it is designed.
    """

    def __init__(self):
        """Initialize a TaskEfficacy instance."""
        super().__init__(
            "TaskEfficacy",
            cleantext(
                """
                The TaskEfficacy property assesses a model's ability
                to correctly perform instances of its task. The means
                of measurement for this property will vary by both
                domain and task. Examples include accuracy, error rate,
                and average precision, but many others are possible.
                """
            ),
        )

    def __repr__(self) -> str:
        """Return the representation needed to reconstruct the object."""
        return f"{self.name}()"



"""
Internal utilities for handling text.
"""

import textwrap


def cleantext(text: str) -> str:
    """
    Clean a text string.

    :param text: The input text
    :type text: str

    :return: The cleaned text
    :rtype: str
    """
    return " ".join(textwrap.dedent(text).strip().split("\n"))



"""
Bind a measurement to one or more properties.
"""

from typing import Union

from ..property import Property
from .measurement import Measurement


def bind(
    measurement: Measurement, *properties: Union[str, Property]
) -> Measurement:
    """
    Bind a measurement to one or more properties.

    :param measurement: The measurement instance to bind
    :type measurement: Measurement
    :param properties: Properties to which the measurement is bound
    :type properties: Union[str, Property]

    :return: The bound measurement instance
    :rtype: Measurement
    """
    if len(properties) == 0:
        raise RuntimeError(
            "Cannot bind measurement to empty collection of properties."
        )
    property_names = (
        property if isinstance(property, str) else property.name
        for property in properties
    )
    measurement._bind(*property_names)
    return measurement



"""
An EvaluationResult instance for a scalar, integral value.
"""

from .evalution_result import EvaluationResult


class Integer(EvaluationResult):
    """
    Integer implements the EvaluationResult
    interface for a single integer value.
    """

    def __init__(self, measurement, value: int):
        """
        Initialize an Integer instance.

        :param measurement: The generating measurement
        :type measurement: Measurement
        :param value: The integer value
        :type value: int
        """
        assert isinstance(value, int), "Argument must be `int`."

        super().__init__(measurement)

        self.value = value
        """The wrapped integer value."""

    def __str__(self) -> str:
        """Return a string representation of the Integer."""
        return f"{self.value}"



"""
Superclass for all measurements.
"""

from __future__ import annotations

import abc
import typing
from typing import Any, List

from .validation import Binding, Bound, Unbound
from .evaluation import EvaluationResult, Opaque
from .validation import ValidationResult, Validator, Ignore


def _has_callable(type, name) -> bool:
    """Determine if `type` has a callable attribute with the given name."""
    return hasattr(type, name) and callable(getattr(type, name))


class Measurement(metaclass=abc.ABCMeta):
    """
    The superclass for all model measurements.
    """

    @classmethod
    def __subclasshook__(cls, subclass):
        """Define the interface for all concrete measurements."""
        return all(_has_callable(subclass, method) for method in ["__call__"])

    def __init__(self, name: str):
        """
        Initialize a new Measurement instance.
        :param name The name of the measurement
        """
        self.name = name
        """The name of the measurement (human-readable identifier)"""

        self.binding: Binding = Unbound()
        """The measurement to property binding."""

        self.validators: List[Validator] = []
        """The collection of measurement validators."""

    @abc.abstractmethod
    @typing.no_type_check
    def __call__(self, *args, **kwargs) -> Any:
        """Evaluate a measurement and return results without semantics."""
        raise NotImplementedError("Cannot evaluate abstract measurement.")

    @typing.no_type_check
    def evaluate(self, *args, **kwargs) -> EvaluationResult:
        """
        Evaluate a measurement and return results with semantics.

        :return: The result of measurement execution, with semantics
        :rtype: EvaluationResult
        """
        data = self(*args, **kwargs)
        return (
            self.semantics(data)
            if hasattr(self, "semantics")
            else Opaque(self, data)
        )

    @typing.no_type_check
    def validate(self, *args, **kwargs) -> List[ValidationResult]:
        """
        Evaluate the measurement and validate results.

        :return: The results of measurement validation
        :rtype: List[ValidationResult]
        """
        result = self.evaluate(*args, **kwargs)
        # Invoke each validator on the result of property evaluation;
        # inject the name of the producing validator and the current
        # binding status for the measurement into the ValidationResult
        return [
            validator(result)
            ._from_data(result)
            ._from_validator(validator)
            ._with_binding(self.binding)
            for validator in self.validators
        ]

    def add_validator(self, validator: Validator):
        """
        Add a validator to the measurement.

        :param validator: The validator instance
        :type validator: Validator
        """
        if any(v.name == "__ignore__" for v in self.validators):
            raise RuntimeError("Cannot add validator for ignored measurement.")
        if any(v.name == validator.name for v in self.validators):
            raise RuntimeError("Validator name must be unique.")
        self.validators.append(validator)

    def with_validator(self, validator: Validator) -> Measurement:
        """
        Add a validator to the measurement.

        :param validator: The validator instance
        :type validator: Validator

        :return: The measurement instance (`self`)
        :rtype: Measurement
        """
        self.add_validator(validator)
        return self

    def ignore(self, reason: str):
        """
        Indicate that measurement validation is ignored.

        :param reason: The reason that measurement validation is ignored
        :type reason: str

        :return: The measurement instance (`self`)
        :rtype: Measurement
        """
        if len(self.validators) > 0:
            raise RuntimeError(
                "Cannot ignore() validation for measurement with validators."
            )

        self.validators.append(
            Validator("__ignore__", lambda _: Ignore(reason))
        )
        return self

    def _bind(self, *properties: str):
        """
        Bind the Measurement to a particular property.

        :param properties Names of the properties to which measurement is bound
        :type properties: str
        """
        if self._is_bound():
            raise RuntimeError("Cannot bind a bound measurement.")
        self.binding = Bound(self, *properties)

    def _is_bound(self) -> bool:
        """
        Determine if the measurement is already bound.

        :return: `True` if the measurement is bound, `False` otherwise
        :rtype: bool
        """
        return bool(self.binding)



"""
An opaque evaluation result, without semantics.
"""

from typing import Dict, Any

from .evalution_result import EvaluationResult


class Opaque(EvaluationResult):
    """
    The 'default' EvaluationResult instance for custom
    measurements that do not define a `semantics()` method.
    """

    def __init__(self, measurement, data: Dict[str, Any]):
        """
        Initialize an Opaque instance.

        :param measurement: The generating measurement
        :type measurement: Measurement
        :param data: The output of the measurement
        :type data: Dict
        """
        super().__init__(measurement)

        self.data = data
        """The raw output from measurement execution."""

    def __getitem__(self, key: str) -> Any:
        """
        Access an item from the wrapped data object.

        :param key: The key that identifies the item to access
        :type key: str

        :return: The value associated with `key`.
        :rtype: Any

        :raises KeyError: If the key is not present
        """
        if key not in self.data:
            raise KeyError(f"Key {key} not found.")
        return self.data[key]

    def __setitem__(self, key: str, value: str) -> None:
        """Raise ValueError to indicate Opaque is read-only."""
        raise ValueError("Opaque is read-only.")



from .storage_cost import StorageCost
from .training_memory_cost import TrainingMemoryCost
from .training_compute_cost import TrainingComputeCost

__all__ = ["StorageCost", "TrainingMemoryCost", "TrainingComputeCost"]



"""
A model test report.

Acknowledgements:
    The contents of a mlte model report, and the
    implementation of model report generation
    in this file is adapted from the work by the
    TensorFlow team in the Model Card Toolkit:
    https://github.com/tensorflow/model-card-toolkit
"""

from __future__ import annotations

import json
import typing
import dataclasses
from dataclasses import dataclass, field
from typing import List, Optional, Dict, Any, Union

from .html import _connected, _generate_html
from ..suite import SuiteReport

from .._private.text import cleantext
from .._private.schema import REPORT_LATEST_SCHEMA_VERSION


class ReportAttribute:
    """The base class for report attributes."""

    def to_json(self) -> Dict[str, Any]:
        """
        Convert a ReportAttribute instance to a JSON document.

        :return: The converted document
        :rtype: Dict[str, Any]
        """
        return dataclasses.asdict(
            self,
            dict_factory=lambda properties: {k: v for k, v in properties if v},
        )


# -----------------------------------------------------------------------------
# Report Sub-Sections
# -----------------------------------------------------------------------------


@dataclass
class Dataset(ReportAttribute):
    """A description of a dataset used to train the model."""

    name: Optional[str] = None
    """An identifier for the dataset."""

    link: Optional[str] = None
    """A method to access the dataset."""

    description: Optional[str] = None
    """A description of the dataset."""

    def __setattr__(self, name: Any, value: Any) -> None:
        """Clean text when adding attributes."""
        value = cleantext(value) if isinstance(value, str) else value
        super().__setattr__(name, value)


@dataclass
class User(ReportAttribute):
    """A description of an intended user of the model."""

    identifier: Optional[str] = None
    """An identifier for the intender user."""

    description: Optional[str] = None
    """A description of the intended user."""

    def __setattr__(self, name: Any, value: Any) -> None:
        """Clean text when adding attributes."""
        value = cleantext(value) if isinstance(value, str) else value
        super().__setattr__(name, value)


@dataclass
class UseCase(ReportAttribute):
    """A description of an intended use case of the model."""

    identifier: Optional[str] = None
    """An identifier for the use case."""

    description: Optional[str] = None
    """A description of the use case."""

    def __setattr__(self, name: Any, value: Any) -> None:
        """Clean text when adding attributes."""
        value = cleantext(value) if isinstance(value, str) else value
        super().__setattr__(name, value)


@dataclass
class Limitation(ReportAttribute):
    """A description of a technical limitation of the model."""

    identifier: Optional[str] = None
    """An identifier for the limitation."""

    description: Optional[str] = None
    """A description of the limitation."""

    def __setattr__(self, name: Any, value: Any) -> None:
        """Clean text when adding attributes."""
        value = cleantext(value) if isinstance(value, str) else value
        super().__setattr__(name, value)


# -----------------------------------------------------------------------------
# Report Sections
# -----------------------------------------------------------------------------


@dataclass
class Metadata(ReportAttribute):
    """Metadata for the report."""

    project_name: Optional[str] = None
    """The name of the project."""

    authors: List[str] = field(default_factory=list)
    """The authors of the report."""

    source_url: Optional[str] = None
    """The URL for model source."""

    artifacts_url: Optional[str] = None
    """The URL for model artifacts."""

    timestamp: Optional[str] = None
    """The time at which the report was generated."""


@dataclass
class ModelDetails(ReportAttribute):
    """The `model details` section of the report."""

    name: Optional[str] = None
    """The name of the model."""

    overview: Optional[str] = None
    """A brief overview of the model."""

    documentation: Optional[str] = None
    """A detailed description of the model."""

    def __setattr__(self, name: Any, value: Any) -> None:
        """Clean text when adding attributes."""
        value = cleantext(value) if isinstance(value, str) else value
        super().__setattr__(name, value)


@dataclass
class ModelSpecification(ReportAttribute):
    """The `model specification` section of the report."""

    domain: Optional[str] = None
    """The domain of the model."""

    architecture: Optional[str] = None
    """A description of the model architecture."""

    input: Optional[str] = None
    """A description of model inputs."""

    output: Optional[str] = None
    """A description of model outputs."""

    data: List[Dataset] = field(default_factory=list)
    """A description of the data used to train the model."""


@dataclass
class Considerations(ReportAttribute):
    """The `considerations` section of the report."""

    users: List[User] = field(default_factory=list)
    """A description of the intended users of the model."""

    use_cases: List[UseCase] = field(default_factory=list)
    """A description of the intended use cases for the model."""

    limitations: List[Limitation] = field(default_factory=list)
    """A description of the technical limitations of the model."""


@dataclass
class Report(ReportAttribute):
    """The top-level model test report."""

    metadata: Metadata = field(default_factory=Metadata)
    """The report metadata."""

    model_details: ModelDetails = field(default_factory=ModelDetails)
    """The model details."""

    model_specification: ModelSpecification = field(
        default_factory=ModelSpecification
    )
    """The model specification."""

    considerations: Considerations = field(default_factory=Considerations)
    """Model considerations."""

    suite: SuiteReport = field(default_factory=lambda: SuiteReport({}))
    """The model test suite report."""

    def _finalize(self) -> Dict[str, Any]:
        """
        Finalize construction of the Report instance.

        :return: Report document
        :rtype: Dict[str, Any]
        """
        # TODO(Kyle): We should perform validation here
        document = dataclasses.asdict(
            self,
            dict_factory=lambda properties: {k: v for k, v in properties if v},
        )
        # Manually serialize the suite-level document
        document["suite"] = self.suite.document
        # Manually insert the schema version
        document["schema_version"] = REPORT_LATEST_SCHEMA_VERSION
        return document

    @typing.no_type_check
    def to_json(
        self, path: Optional[str] = None
    ) -> Union[Dict[str, Any], None]:
        """
        Convert the Report to a JSON document.

        The converted document is saved to the provided `path`
        if it is specified, otherwise it is returned as a Python
        dictionary.

        :param path: The path to which document is saved
        :type path: Optional[str]

        :return: Python-dict representation of JSON document, or None
        :rtype: Union[Dict[str, Any], None]
        """
        document = self._finalize()
        if path is None:
            return document

        with open(path, "w") as f:
            json.dump(document, f, indent=2)

        return None

    def to_html(self, path: Optional[str] = None) -> Union[str, None]:
        """
        Convert the Report to an HTML document.

        The converted document is saved to the provided `path`
        if it is specified, otherwise it is returned as a string.

        :param path: The path to which document is saved
        :type path: Optional[str]

        :return: String representation of HTML document, or None
        :rtype: Union[str, None]

        :raises RuntimeError: If network connection is unavailable
        """
        if not _connected():
            raise RuntimeError(
                "HTML report generation requires a network connection."
            )

        # Generate the string representation of HTML document
        html = _generate_html(self._finalize())

        if path is not None:
            with open(path, "w") as f:
                f.write(html)

        return html if path is None else None



"""
Utilities for HTML report generation.
"""

import json
import socket
import tempfile
from urllib import request
from typing import Dict, Any

# The endpoint for resolving endpoints for report generation
RESOLUTION_ENDPOINT = "https://raw.githubusercontent.com/mlte-team/mlte/master/assets/endpoints.txt"  # noqa


def _connected(host: str = "8.8.8.8", port: int = 53, timeout: int = 2) -> bool:
    """
    Determine if internet connectivity is available.

    :param host: The host used to test connectivity
    :type host: str
    :param port: The port used to test connectivity
    :type port: int
    :param timeout: The connection timeout
    :type timeout: int

    :return `True` if connected to the internet, `False` otherwise
    :rtype: bool
    """
    try:
        socket.setdefaulttimeout(timeout)
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
        return True
    except socket.error:
        return False


def _resolve_endpoint(meta_endpoint: str = RESOLUTION_ENDPOINT) -> str:
    """
    Resolve the endpoint for report generation.

    :param meta_endpoint: The endpoint for resolution requests
    :type meta_endpoint: str

    :return: A report generation endpoint
    :rtype: str

    :raises RuntimeError: If unable to resolve endpoint
    """
    with tempfile.NamedTemporaryFile() as f:
        path = f.name
        try:
            request.urlretrieve(meta_endpoint, path)
        except Exception:
            raise RuntimeError(
                "Unable to resolve endpoint for report generation."
            )

        with open(path, "r") as endpoints:
            # Return an arbitrary line from the list of endpoints
            for line in endpoints:
                return line.strip()

        raise RuntimeError("Unreachable")


def _generate_html(document: Dict[str, Any]) -> str:
    """
    Generate the HTML representation of a Report.

    :param document: The JSON document representation of the report
    :type document: Dict[str, Any]

    :return: The HTML representation of the report, as a string
    :rtype: str

    :raises RuntimeError: If report generation fails
    """
    assert _connected(), "Broken precondition."

    # Resolve the endpoint for report generation
    endpoint = _resolve_endpoint()

    # Construct the request with the report document
    req = request.Request(
        endpoint,
        method="POST",
        data=json.dumps(document).encode("utf-8"),
    )
    req.add_header("Content-Type", "application/json; charset=utf-8")

    with request.urlopen(req) as response:
        # TODO(Kyle): Better error handling.
        if response.status != 200:
            raise RuntimeError("Request failed.")
        return str(response.read().decode("utf-8"))



"""
TrainingMemoryCost property definition.
"""

from ..property import Property
from ..._private.text import cleantext


class TrainingMemoryCost(Property):
    """
    The TrainingMemoryCost property reflects the cost
    of model training associated with memory resources.
    """

    def __init__(self):
        """
        Initialize a TrainingMemoryCost instance.
        """
        super().__init__(
            "TrainingMemoryCost",
            cleantext(
                """
                The TrainingMemoryCost property assesses the
                memory requirements of model training. This might
                be measured by the memory requirements of training
                processes that run locally, or the cost of memory
                resources required for training processes that run
                on on-demand cloud infrastructure.
                """
            ),
        )

    def __repr__(self) -> str:
        """Return the representation needed to reconstruct the object."""
        return f"{self.name}()"



"""
CPU utilization measurement for local training processes.
"""

import time
import subprocess
from typing import Dict, Any
from subprocess import SubprocessError

from ..measurement import Measurement
from ..evaluation import EvaluationResult
from ..validation import Validator, Success, Failure
from ..._private.platform import is_windows


class CPUStatistics(EvaluationResult):
    """
    The CPUStatistics class encapsulates data
    and functionality for tracking and updating
    CPU consumption statistics for a running process.
    """

    def __init__(
        self, measurement: Measurement, avg: float, min: float, max: float
    ):
        """
        Initialize a CPUStatistics instance.

        :param measurement: The generating measurement
        :type measurement: Measurement
        :param avg: The average utilization
        :type avg: float
        :param min: The minimum utilization
        :type min: float
        :param max: The maximum utilization
        :type max: float
        """
        super().__init__(measurement)

        self.avg = avg
        """The average CPU utilization, as a proportion."""

        self.min = min
        """The minimum CPU utilization, as a proportion."""

        self.max = max
        """The maximum CPU utilization, as a proportion."""

    def __str__(self) -> str:
        """Return a string representation of CPUStatistics."""
        s = ""
        s += f"Average: {self.avg:.2f}%\n"
        s += f"Minimum: {self.min:.2f}%\n"
        s += f"Maximum: {self.max:.2f}%"
        return s


def _get_cpu_usage(pid: int) -> float:
    """
    Get the current CPU usage for the process with `pid`.

    :param pid: The identifier of the process
    :type pid: int

    :return: The current CPU utilization as percentage
    :rtype: float
    """
    try:
        stdout = subprocess.check_output(
            ["ps", "-p", f"{pid}", "-o", "%cpu"], stderr=subprocess.DEVNULL
        ).decode("utf-8")
        return float(stdout.strip().split("\n")[1].strip())
    except SubprocessError:
        return -1.0
    except ValueError:
        return -1.0


class LocalProcessCPUUtilization(Measurement):
    """Measures CPU utilization for a local process."""

    def __init__(self):
        """Initialize a new LocalProcessCPUUtilization measurement."""
        super().__init__("LocalProcessCPUUtilization")
        if is_windows():
            raise RuntimeError(
                f"Measurement {self.name} is not supported on Windows."
            )

    def __call__(self, pid: int, poll_interval: int = 1) -> Dict[str, Any]:
        """
        Monitor the CPU utilization of process at `pid` until exit.

        :param pid: The process identifier
        :type pid: int
        :param poll_interval: The poll interval in seconds
        :type poll_interval: int

        :return: The collection of CPU usage statistics
        :rtype: Dict
        """
        stats = []
        while True:
            util = _get_cpu_usage(pid)
            if util < 0.0:
                break
            stats.append(util / 100.0)
            time.sleep(poll_interval)

        return {
            "avg_utilization": sum(stats) / len(stats),
            "min_utilization": min(stats),
            "max_utilization": max(stats),
        }

    def semantics(self, data: Dict[str, Any]) -> CPUStatistics:
        """
        Provide semantics for measurement output.

        :param data: Measurement output data
        :type data: Dict

        :return: CPU utilization statistics
        :rtype: CPUStatistics
        """
        assert "avg_utilization" in data, "Broken invariant."
        assert "min_utilization" in data, "Broken invariant."
        assert "max_utilization" in data, "Broken invariant."
        return CPUStatistics(
            self,
            avg=data["avg_utilization"],
            min=data["min_utilization"],
            max=data["max_utilization"],
        )

    def with_validator_max_utilization_not_greater_than(
        self, threshold: float
    ) -> Measurement:
        """
        Add a validator for maximum CPU utilization.

        :param threshold: The threshold value for maximum utilization
        :type threshold: float

        :return: The measurement instance (`self`)
        :rtype: Measurement
        """
        return self.with_validator(
            Validator(
                "MaximumUtilization",
                lambda stats: Success(
                    f"Maximum utilization {stats.max:.2f} "
                    f"below threshold {threshold:.2f}"
                )
                if stats.max <= threshold
                else Failure(
                    (
                        f"Maximum utilization {stats.max:.2f} "
                        f"exceeds threshold {threshold:.2f}"
                    )
                ),
            )
        )

    def with_validator_avg_utilization_not_greater_than(
        self, threshold: float
    ) -> Measurement:
        """
        Add a validator for average CPU utilization.

        :param threshold: The threshold value for average utilization
        :type threshold: float

        :return: The measurement instance (`self`)
        :rtype: Measurement
        """
        return self.with_validator(
            Validator(
                "AverageUtilization",
                lambda stats: Success(
                    f"Average utilization {stats.max:.2f} "
                    f"below threshold {threshold:.2f}"
                )
                if stats.avg <= threshold
                else Failure(
                    (
                        f"Average utilization {stats.avg:.2f} "
                        "exceeds threshold {threshold:.2f}"
                    )
                ),
            )
        )



"""
Internal utilities for schema management.

Acknowledgements:
    The design of schema versioning in this project,
    and the implementation of schema file resolution
    in this file is adapted from the work by the
    TensorFlow team in the Model Card Toolkit:
    https://github.com/tensorflow/model-card-toolkit
"""

import os
import json
import pkgutil
import jsonschema
from typing import Any, Dict, Optional

# The identifier for the latest schema for suites
SUITE_LATEST_SCHEMA_VERSION = "0.0.1"
# The identifier for the latest schema for reports
REPORT_LATEST_SCHEMA_VERSION = "0.0.1"

# Version identifiers for suite schemas
_SUITE_SCHEMA_VERSIONS = frozenset(("0.0.1",))
# Version identifier for report schemas
_REPORT_SCHEMA_VERSIONS = frozenset(("0.0.1",))

# The name of the schema file for all schemas
_SCHEMA_FILE_NAME = "schema.json"

# -----------------------------------------------------------------------------
# Schema Validation
# -----------------------------------------------------------------------------


def validate_suite_schema(
    document: Dict[str, Any], version: Optional[str] = None
):
    """
    Validate the schema of a suite output document.

    :param document: The document instance
    :type document: Dict[str, Any]
    :param version: The identifier for the schema version
    :type version: Optional[str]

    :raises ValueError: If an invalid schema is specified
    :raises ValidationError: If the instance fails validation
    """
    version = (
        version or document.get("schema_version") or SUITE_LATEST_SCHEMA_VERSION
    )
    jsonschema.validate(instance=document, schema=_find_suite_schema(version))


def validate_report_schema(
    document: Dict[str, Any], version: Optional[str] = None
):
    """
    Validate the schema of a suite output document.

    :param document: The document instance
    :type document: Dict[str, Any]
    :param version: The identifier for the schema version
    :type version: Optional[str]

    :raises ValueError: If an invalid schema is specified
    :raises RuntimeError: If the instance fails validation
    """
    version = (
        version
        or document.get("schema_version")
        or REPORT_LATEST_SCHEMA_VERSION
    )
    jsonschema.validate(instance=document, schema=_find_report_schema(version))


# -----------------------------------------------------------------------------
# Schema Resolution
# -----------------------------------------------------------------------------


def _find_schema(version: str, subdirectory: str) -> Dict[str, Any]:
    """
    Find, load, and return the JSON schema.

    :param version: The version identifier for the schema
    :type version: str
    :param subdirectory: The name of the subdirectory to perform search
    :type subdirectory: str

    :return: The loaded schema
    :rtype: Dict[str, Any]

    :raises ValueError: If an invalid schema is specified
    """
    # Construct the path to the schema file
    path = os.path.join(
        "schema", subdirectory, f"v{version}", _SCHEMA_FILE_NAME
    )
    # Locate the artifact, load, and return it
    data = pkgutil.get_data("mlte", path)
    if data is None:
        raise ValueError(f"Schema {path} not found.")
    return json.loads(data)  # type: ignore


def _find_suite_schema(version: Optional[str] = None) -> Dict[str, Any]:
    """
    Find, load, and return the JSON schema for suite output.

    :param version: The version identifier for the schema
    :type version: Optional[str]

    :return: The loaded schema
    :rtype: Dict[str, Any]

    :raises ValueError: If an invalid schema is specified
    """
    if version is None:
        version = SUITE_LATEST_SCHEMA_VERSION
    if version not in _SUITE_SCHEMA_VERSIONS:
        raise ValueError(f"Invalid suite schema version {version} specified.")
    return _find_schema(version, "suite")


def _find_report_schema(version: Optional[str] = None):
    """
    Find, load, and return the JSON schema for report output.

    :param version: The version identifier for the schema
    :type version: Optional[str]

    :return: The loaded schema
    :rtype: Dict[str, Any]

    :raises ValueError: If an invalid schema is specified
    """
    if version is None:
        version = REPORT_LATEST_SCHEMA_VERSION
    if version not in _REPORT_SCHEMA_VERSIONS:
        raise ValueError(f"Invalid report schema version {version} specified.")
    return _find_schema(version, "report")



"""
Platform discovery.
"""

import os


def is_windows() -> bool:
    """
    Determine if the platform is Windows.

    :return: `True` if the OS is Windows, `False` otherwise.
    :rtype: bool
    """
    return os.name == "nt"


def is_nix() -> bool:
    """
    Determine if the platform is UNIX-like.

    :return: `True` if the OS is UNIX-like, `False` otherwise.
    :rtype: bool
    """
    return os.name == "posix"



"""
StorageCost property definition.
"""

from ..property import Property
from ..._private.text import cleantext


class StorageCost(Property):
    """
    The StorageCost property reflects the cost of model storage.
    """

    def __init__(self):
        """Initialize a StorageCost instance."""
        super().__init__(
            "StorageCost",
            cleantext(
                """
                The StorageCost property assesses the storage requirements of a
                model. These requirements may be expressed in a variety of ways,
                including the physical size of the model when it is persisted
                to stable storage, or the number of parameters it contains.
                """
            ),
        )

    def __repr__(self) -> str:
        """Return the representation needed to reconstruct the object."""
        return f"{self.name}()"



"""
Utilities related to measurement collection.
"""

from typing import Union, Any, Iterable, List


def flatten(*collections: Union[Any, Iterable[Any]]) -> List[Any]:
    """
    Flatten a variadic number of collections.

    :param collections: The collections to flatten
    :type collections: Union[Any, Iterable[Any]]

    :return: The flattened collection
    :rtype: List[Any]
    """
    clone = [collection for collection in collections]
    for i, _ in enumerate(clone):
        while i < len(clone) and isinstance(clone[i], (list, tuple)):
            clone[i : i + 1] = clone[i]  # noqa
    return clone



from .validation_result import (
    ValidationResult,
    Success,
    Failure,
    Ignore,
    Binding,
    Bound,
    Unbound,
)
from .validator import Validator

__all__ = [
    "ValidationResult",
    "Success",
    "Failure",
    "Ignore",
    "Validator",
    "Binding",
    "Bound",
    "Unbound",
]



"""
Indicates the outcome of measurement evaluation.
"""

import abc


class EvaluationResult(metaclass=abc.ABCMeta):
    """
    The EvaluationResult class serves as the base class of
    all semantically-enriched measurement evaluation results.
    The EvaluationResult provides a common interface for
    inspecting the results of measurement evaluation, and also
    encapsulates the functionality required to uniquely
    associate evaluation results with the originating measurement.
    """

    def __init__(self, measurement):
        """
        Initialize an EvaluationResult instance.

        :param measurement: The generating measurement
        :type measurement: Measurement
        """
        # Store the name of the generating measurement
        self.measurement_name = measurement.name



from .local_process_cpu_utilization import (
    LocalProcessCPUUtilization,
    CPUStatistics,
)

# TODO(Kyle): Find a more elegant way to express these exports
__all__ = ["LocalProcessCPUUtilization", "CPUStatistics"]
